"""
The reason this isn't just combined in issues.py is that in theory a service can
depend on qai without needing spaCy

so only if they import this module will we import from spaCy
"""
import traceback
from typing import Optional

from qai.issues import make_issue

try:
    from spacy.tokens.span import Span
except ModuleNotFoundError as e:
    print("If you want to use SpacyFactor, please install spaCy >=2.1.0")
    raise e


class SpacyFactor:
    def __init__(
        self,
        issue_type: str,
        simple_description: str,
        description: Optional[str] = None,
        debug=False,
    ):
        self.issue_type = issue_type
        self.simple_description = simple_description
        if description:
            self.description = description
        self.debug = debug

    @staticmethod
    def match_case(span: Span, suggestions):
        """
        suggestions respect original casing
        ex. purchase -> buy (lower)
        ex. PURCHASE -> BUY (upper)
        ex. Purchase -> Buy (title)
        """
        span_text = span.text
        try:
            if len(span_text.strip()):
                words = span_text.split(" ")
                is_sent_start = len(span) and span[0].is_sent_start
                if words[0].islower():
                    # this should exclude I
                    # todo - capitalized proper nouns
                    suggestions = [
                        " ".join(
                            [
                                word.lower() if word != "I" else word
                                for word in sug.split()
                            ]
                        )
                        for sug in suggestions
                    ]
                elif len(words[0]) > 1 and words[0].isupper():
                    # len condition is to exclude capitalization caused by "I"
                    suggestions = [s.upper() for s in suggestions]
                elif (len(words[0]) > 1 and words[0][0].isupper()) or (
                    is_sent_start and len(span[0]) == 1
                ):
                    # title case - first letter check & conversion due to contractions
                    for i, s in enumerate(suggestions):
                        if len(s.strip()):
                            s_words = s.split(" ")
                            suggestions[i] = " ".join(
                                [s_words[0][0].upper() + s_words[0][1:]] + s_words[1:]
                            ).strip()
        except:
            suggestions_log = ",".join(suggestions)
            print(traceback.format_exc())
            print(f"Case matching failed for span: {span_text} and {suggestions_log}")

        return suggestions

    def __call__(
        self, spacy_span: Span, meta_subcategory=None, describer=None, cased=True,
    ):
        """
        convenient wrapper around make_issue if you are using spaCy

        usage example:

        ```python
        from spacy.tokens import Span
        from app.factor import SpacyFactor


        SOV = SpacyFactor(
            "subject_object_verb_spacing",
            "Keep the subject, verb, and object of a sentence close together to help the reader understand the sentence."
        )

        Span.set_extension("score", default=0)
        Span.set_extension("suggestions", default=[])

        doc = nlp("Holders of the Class A and Class B-1 certificates will be entitled to receive on each Payment Date, to the extent monies are available therefor (but not more than the Class A Certificate Balance or Class B-1 Certificate Balance then outstanding), a distribution.")
        score = analyze(doc)
        if score is not None:
            span = Span(doc, 0, len(doc))  # or whichever TOKENS are the issue (don't have to worry about character indexes)
            span._.score = score
            span._.suggestions = get_suggestions(doc)
            issues = SOV(span)
        ```
        """
        text, start, end = spacy_span.text, spacy_span.start_char, spacy_span.end_char
        score = spacy_span._.score if spacy_span.has_extension("score") else 0
        suggestions = (
            spacy_span._.suggestions if spacy_span.has_extension("suggestions") else []
        )

        learn_more = (
            spacy_span._.learn_more if spacy_span.has_extension("learn_more") else ""
        )
        paragraph = (
            spacy_span._.paragraph if spacy_span.has_extension("paragraph") else False
        )
        header = spacy_span._.header if spacy_span.has_extension("header") else ""
        meta_dict = (
            spacy_span._.meta_dict if spacy_span.has_extension("meta_dict") else {}
        )
        visible = spacy_span._.visible if spacy_span.has_extension("visible") else True
        # SIMPLE DESCRIPTION IS DIFFERENT, since it is set when initialized
        # so we need to check that the span doesn't just have an empty value
        simple_description = (
            spacy_span._.simple_description
            if (
                spacy_span.has_extension("simple_description")
                and spacy_span._.simple_description != ""
            )
            else self.simple_description
        )
        accept_all_changes = (
            spacy_span._.accept_all_changes
            if spacy_span.has_extension("accept_all_changes")
            else False
        )

        if cased:
            suggestions = self.match_case(spacy_span, suggestions)

        if describer:
            description = describer(spacy_span)
        else:
            description = self.description

        if meta_subcategory:
            subcategory = meta_subcategory
        else:
            subcategory = ""

        i = make_issue(
            segment=spacy_span.doc.text,
            start=start,
            end=end,
            issue_type=self.issue_type,
            score=score,
            description=description,
            simpleDescription=simple_description,
            suggestions=suggestions,
            subCategory=subcategory,
            learnMore=learn_more,
            paragraph=paragraph,
            header=header,
            metaDict=meta_dict,
            accept_all_changes=accept_all_changes,
            debug=self.debug,
            visible=visible,
        )
        return i



import time
from collections import namedtuple
from importlib.metadata import version
from typing import Dict, Optional

import sanic
import ujson
from sanic import Sanic
from sanic.log import logger
from sanic.request import Request
from sanic.response import HTTPResponse

from qai.util import convert_issue_spans_to_utf16, get_cpu_quota_within_docker
from qai.validation import Validator, filter_no_ops, fix_extra_space

SegmentWrapper = namedtuple("SegmentWrapper", ["id", "segment", "meta", "category"])


async def set_valid_segments(request: Request):
    """
    uses `request.ctx.validator`, which must already be set,
    to validate segments.
    Stores valid segments to `request.ctx.segments` and also sets `request.ctx.info`
    This is used by default in QRest, or can be used in a standalone sanic app with
    ```python
    from qai.server import set_valid_segments
    ...
    app.request_middleware.append(set_valid_segments)
    ```
    """
    if request.method != "POST":
        return None
    data = request.json
    req = data.get("request", {"segments": []})
    category = req.get("category")
    # store category in context
    request.ctx.category = category
    segs = req.get("segments", [])
    # store segments in context
    request.ctx.segments = []
    info = {}
    for el in segs:
        s_id = el["content"]["segmentId"]
        segment = el["content"]["segment"]
        meta = el.get("meta", {})
        # store extra info if eventually needed
        info[s_id] = {k: v for k, v in req.items() if k != "segments"}
        is_valid = request.app.ctx.validator(segment)
        tpe = el["content"]["tpe"]
        if is_valid:
            request.ctx.segments.append(SegmentWrapper(s_id, segment, meta, category))
            info[s_id].update({"type": tpe, "segmentId": s_id, "category": category})
            # jam all non-meta into request.ctx.info
    request.ctx.info = info


async def postprocess_issues(
    request: Request, response: HTTPResponse
) -> Optional[HTTPResponse]:
    """
    THIS MUST BE THE LAST MIDDLEWARE
    since when it works, it returns, which bypasses remaining middleware

    It changes the string indexing to be utf-16 based, and filters no ops
    This is used by default in QRest, or can be used in a standalone sanic app with

    ```python
    from qai.server import postprocess_issues

    ...
    app.response_middleware.append(postprocess_issues)
    ```

    However, if you use your own handler and import this middleware, you must set
    the key "segment" in addition to the 3 standard keys of a response,
    "category", "segmentId", "issues"

    if there is an error, though, it should still respond, since when there is
    an exception in middleware, sanic sends whatever the handler returned
    which in this case is issues which possibly contain no-ops and have the wrong indexing
    """
    if request.method != "POST":
        # don't run on health checks
        return None
    res = ujson.loads(response.body)
    for segment_response in res:
        segment = segment_response["segment"]
        segment_response["issues"] = filter_no_ops(segment, segment_response["issues"])
        if request.app.ctx.fix_extra_space:
            segment_response["issues"] = fix_extra_space(
                segment, segment_response["issues"]
            )
        del segment_response["segment"]
        if not request.app.ctx.utf16_convert:
            continue
        for i, issue in enumerate(segment_response["issues"]):
            segment_response["issues"][i] = convert_issue_spans_to_utf16(segment, issue)
    # since we return, no other middleware will be run after this
    return sanic.response.json(res)


async def main_handler(request: Request) -> HTTPResponse:
    """
    The default handler, which works when the service responds with one category

    Must come after the middleware which sets request.ctx.analyzer

    If you have to respond with multiple categories it should still work if you
    pass QRest(..., issue_type_to_categories={"some_type": "some_category"})

    Most of the code is just for handling issue_type to category mapping
    """
    it2c = request.app.ctx.issue_type_to_category
    res = []
    for s_id, segment, meta, category in request.ctx.segments:
        issues = request.app.ctx.analyzer(segment, meta, request.ctx.info[s_id])
        if it2c is None:
            segment_response = {
                "category": category,
                "segmentId": s_id,
                "segment": segment,
                "issues": issues,
            }
            res.append(segment_response)
        else:
            types = it2c.keys()
            for t in types:
                res.append(
                    {
                        "category": it2c[t],
                        "segmentId": s_id,
                        "segment": segment,
                        "issues": [i for i in issues if i["issueType"] == t],
                    }
                )
    return sanic.response.json(res)


async def get_root(request: Request) -> HTTPResponse:
    return sanic.response.json(request.app.ctx.status_dump)


async def add_start_time(request: Request):
    """Prepend initial time when this request was served."""
    request.ctx.start_time = time.time()


async def add_perf_stats(request: Request, response: HTTPResponse):
    """
    Prepend initial time when this request was served.
    Log the access long on each request.
    Log num of segments, sentences, words in payload.
    Log sentence and word speed.
    """
    is_get = request.method.lower() == "get"
    is_check = request.path in ["/ready", "/healthy"]
    should_log = not (is_get or is_check)
    if not should_log:
        return
    latency = round((time.time() - request.ctx.start_time) * 1000)
    sids = "[" + ", ".join([s.id for s in request.ctx.segments]) + "]"
    logger.info(
        (
            "{method} {url} {status} {latency}ms {reqbytes}reqBytes {resbytes}resBytes"
            " segIds={segmentIds}"
        ).format(
            method=request.method,
            url=request.url,
            status=response.status,
            latency=latency,
            reqbytes=len(request.body),
            resbytes=len(response.body),
            segmentIds=sids,
        )
    )


class QRest:
    def __init__(
        self,
        analyzer,
        category: str,
        host: str = "0.0.0.0",
        port: int = 5000,
        workers: Optional[int] = None,
        debug: bool = False,
        ignore_html: bool = True,
        ignore_token_fraction: float = 0.5,
        app: Optional[Sanic] = None,
        issue_type_to_category: Dict[str, str] = None,
        utf16_convert: bool = True,
        fix_extra_space: bool = True,
    ):
        """
        A convenience class for building a server class given an Analyzer class

        you can control parameters of the server with host/port/workers/debug
        or, you can pass a Sanic app using the app= kwarg

        If your service handles multiple categories (like punctuation), you must either build the app yourself
        or pass issue_type_to_category, a Dict that maps issue type (str) to category (str)

        Params:
            analyzer - a callable class with signature (segment, meta=None, extra_info=None) -> issues: List[dict]
            category: str - the category to use if not using `issue_type_to_category`
            host: str = "0.0.0.0" - the address to serve from
            port: int = 5000 - the port to serve on
            workers: Optional[int] = None - how many workers to use. If `None`, uses all available cores, respecting k8s limits
            debug: bool = False - should the server run in debug mode
            ignore_html: bool = True - should we skip processing segments with `<.*>`
            ignore_token_fraction: float = 0.5 - portion of tokens which, if they are nonalpha, we ignore the segment
            app: Optional[Sanic] = None - a Sanic server with routes set already. We still add middleware
            issue_type_to_category: Dict[str, str] = None - if set, separates a single dict with different issue types into multiple dicts
            utf16_convert: bool = True - should we convert string indices to utf16, instead of python's version
            fix_extra_space: bool = True - should we change issue boundaries if deletion causes an extra space?
                usually the right thing to do, but conceivably could cause problems, so you can disable it
        """
        self.host = host
        self.port = port
        self.category = category
        self.debug = debug
        self.ignore_html = ignore_html
        self.validator = Validator(
            ignore_html=ignore_html, ignore_token_fraction=ignore_token_fraction,
        )
        self.analyzer = analyzer
        if not callable(analyzer):
            raise ValueError(
                "Analyzer instance must be callable, with signature (segment, meta=None, extra_info=None)."
            )

        if workers is None:
            self.workers = get_cpu_quota_within_docker()
        else:
            self.workers = int(workers)

        if app is None:
            app = Sanic(__name__)
            app.ctx.issue_type_to_category = issue_type_to_category
            # now the handlers will get called
            app.add_route(main_handler, "/", methods=["POST"])
            app.add_route(main_handler, "/v2", methods=["POST"])
        # I assume even custom-passed apps will want these attached
        app.ctx.validator = self.validator
        app.ctx.analyzer = self.analyzer
        app.ctx.status_dump = {
            "service": self.category,
            "status": "up",
            "host": self.host,
            "port": self.port,
            "qai version": version("qai"),
        }
        app.ctx.utf16_convert = utf16_convert
        app.ctx.fix_extra_space = fix_extra_space
        # validate segments, attach valid ones to request.ctx
        app.request_middleware.append(set_valid_segments)
        # Add the request and response middleware for logging
        app.request_middleware.append(add_start_time)
        app.response_middleware.append(add_perf_stats)
        # Handlers Run Here
        # middleware to filter no-ops
        # THIS MUST BE THE LAST MIDDLEWARE
        app.response_middleware.append(postprocess_issues)
        # attach and add health check, status routes
        self.app = app
        self.add_simple_routes()

    def add_simple_routes(self):
        self.app.add_route(get_root, "/", methods=["GET"])
        self.app.add_route(get_root, "/ready", methods=["GET"])
        self.app.add_route(get_root, "/healthy", methods=["GET"])

    def get_future(self):
        """
        Return a Future (Promise in JS land) that can be put on an event loop

        You probably don't need this unless you are doing crazy stuff
        """
        return self.app.create_server(host=self.host, port=self.port)

    def connect(self):
        """
        Doesn't return, makes a blocking connection
        Only use if you are ONLY using REST
        This is a more robust REST server than get_future makes
        see:
        https://sanic.readthedocs.io/en/latest/sanic/deploying.html#asynchronous-support
        """
        return self.app.run(
            host=self.host,
            access_log=False,
            port=self.port,
            workers=self.workers,
            debug=self.debug,
        )



import itertools
import math
import multiprocessing
from pathlib import Path
from typing import Any, Dict


def _ucp_to_utf16_charmap(s: str):
    """
    mostly copied from
    https://stackoverflow.com/questions/56280011/keeping-java-string-offsets-with-unicode-consistent-in-python
    converts from python indices (unicode code points) to indices
    """
    encoding = "UTF-16LE"
    chrLengths = [len(bytearray(ch, encoding)) // 2 for ch in s]
    utf16indices = [0] + list(itertools.accumulate(chrLengths))
    return utf16indices


def convert_issue_spans_to_utf16(s: str, issue: Dict[str, Any]) -> Dict[str, Any]:
    # modifies in place but also returns
    charmap = _ucp_to_utf16_charmap(s)
    f = charmap[issue["from"]]
    u = charmap[issue["until"]]
    # store these if needed for python string manipulation
    issue["_from_p"] = issue["from"]
    issue["_until_p"] = issue["until"]
    # but set the public values to utf-16
    issue["from"] = f
    issue["until"] = u
    return issue


def get_cpu_quota_within_docker():
    """
    By default, we use mp.cpu_count()
    HOWEVER, if there are cpu limits in certain paths, we assume those are
    from k8s/docker and override with those values
    """
    cpu_cores = multiprocessing.cpu_count()

    cpu_shares = Path("/sys/fs/cgroup/cpu/cpu.shares")

    if cpu_shares.exists():
        with cpu_shares.open("rb") as r:
            request_cpu_shares = int(r.read().strip())
            cpu_cores = (
                math.ceil(request_cpu_shares / 1024)
                if request_cpu_shares > 0
                else multiprocessing.cpu_count()
            )
            print(
                f"CPU request limit: {round(request_cpu_shares / 1024, 2)}, will spin {cpu_cores} worker(s)."
            )
    else:
        print(f"CPU_shares not found, will spin {cpu_cores} worker(s).")
    return cpu_cores



from typing import Any, Dict, List, Union


def make_issue(
    segment: str,
    start: int,
    end: int,
    issue_type: str,
    description: str,
    simpleDescription: str,
    suggestions: List[str] = [],
    subCategory: str = "",
    learnMore: str = "",
    paragraph: bool = False,
    header: str = "",
    metaDict: Dict[str, Any] = {},
    accept_all_changes: bool = False,
    score: Union[float, int] = 0,
    debug: bool = False,
    visible: bool = True,
) -> Dict[str, Any]:
    """
    Turn a bunch of parameters into Writer's dictionary issue format
    old reference:
    https://qordoba.atlassian.net/wiki/spaces/HOME/pages/840368141/Content-AI

    update as of ~May 2021: we are now using v2 issues only,
    and should default to utf16 string indexing, but allow python-unicode
    string-indexing just for debugging

    WARNING this now takes `segment` as the first parameter.
    YOU HAVE TO PASS THE WHOLE SEGMENT to get the utf16 indexes right
    if you do not want to convert to utf16 indices (for debugging):
        - you still need to pass the whole segment, sorry
        - pass make_issue(..., unicode_convert=None)
    """
    issue = {
        "issueType": issue_type,
        "from": start,
        "until": end,
        "score": score,
        "suggestions": suggestions,
        "description": description,
        "simpleDescription": simpleDescription,
        "meta": {"paragraph": paragraph},
        "visible": visible,
    }
    # for Doris/analytics, who can see the meta but not the visible field
    issue["meta"]["visible"] = visible

    if subCategory:
        issue["meta"]["subCategory"] = subCategory
    if learnMore:
        issue["meta"]["learnMore"] = learnMore
    if header:
        issue["meta"]["header"] = header
    if accept_all_changes:
        issue["meta"]["acceptAllChanges"] = accept_all_changes

    if debug:
        # log warnings if meta_dict overwrites issue["meta"]
        common = [value for value in metaDict.keys() if value in issue["meta"]]
        for c in common:
            if metaDict[c] != issue["meta"][c]:
                print(
                    f"Meta_dict {c}:{metaDict[c]} is overwriting issue meta {c}:{issue['meta'][c]}"
                )
        # also generate new_segment
        suggestion = suggestions[0] if len(suggestions) else ""
        issue["new_segment"] = segment[:start] + suggestion + segment[end:]
    # this has to come after the if debug warning
    issue["meta"].update(metaDict)
    return issue



import json as __mod_json
import os
import pprint

import requests


__metas = {
    "gec": {},
    "hate-speech": {},
    "gender-nouns": {},
    "gender-pronouns": {},
    "infinitive": {},
    "healthy-communication": {},
    "confidence": {},
    "parallelism": {},
    "paragraph": {
        "sentenceComplexity": 8,
        "vocabulary": 8,
        "paragraphLength": "medium",
    },
    "punctuation": {"punctuation": True, "oxfordComma": "use", "emDashSpaces": "use",},
    "casing": {"useSentenceCase": "use"},
    "acronym": {"commonAcronymCheck": True},
    "sensitivity": {
        "age": True,
        "disability": True,
        "genderIdentity": True,
        "raceEthnicityNationality": True,
        "sexualOrientation": True,
        "substanceUse": True,
    },
    # inclusivity/sensitivity are the same
    "inclusivity": {
        "age": True,
        "disability": True,
        "genderIdentity": True,
        "raceEthnicityNationality": True,
        "sexualOrientation": True,
        "substanceUse": True,
    },
    "sentence-complexity": {
        "debug": True,
        "sentenceComplexity": 8,
        "vocabulary": 8,
        "paragraphLength": "medium",
    },
    "readability": {
        "grade": 4,
        "index": "flesch-kincaid-grade-level",
        "ignore": ["financial"],
    },
    "plain-language": {
        "passiveVoice": True,
        "wordiness": True,
        "unclearReferences": True,
    },
    "vocabulary": {
        "debug": True,
        "sentenceComplexity": 8,
        "vocabulary": 8,
        "paragraphLength": "medium",
    },
}


def __build_qai_payload(
    segment: str, meta: dict = {}, category: str = "TEST-CATEGORY"
) -> dict:
    qai_payload = {
        "request": {
            "organizationId": 1,
            "workspaceId": 1,
            "personaId": 94,
            "documentId": "61845063-809a-4398-b741-b841c77c6676",
            "version": 1,
            "userId": 1,
            "clientId": 1,
            "segments": [
                {
                    "category": category,
                    "content": {
                        "segmentId": "95ea2705-6a0d-48c0-9eeb-78dcf69cef79",
                        "tpe": "document",
                        "segment": "segment",
                        "from": 10,
                    },
                    "meta": meta,
                }
            ],
        }
    }
    qai_payload["request"]["segments"][0]["content"]["segment"] = segment
    qai_payload["request"]["segments"][0]["meta"] = meta
    return qai_payload


def test_qai(segment: str, meta: dict = {}, url: str = "http://localhost:5000"):
    """for testing a qai service by URL
    if it is an established service, you can use like
    test_qai("segment to test", "gec", 8080)
    where `"gec"` tells it what meta to use
    and 8080 means localhost port 8080

    if this is a new service, or you want to specify the meta, then use

    test_qai("segment to test", {"something": 10}, 5000)

    for testing via k8s proxy, use name_test_qai
    """
    if isinstance(meta, str):
        meta = __metas.get(meta, {})
    qai_payload = __build_qai_payload(segment, meta)
    if isinstance(url, int):
        url = f"http://localhost:{url}"
    response = requests.post(url, json=qai_payload)
    try:
        return response.json()
    except Exception as e:
        print("something went wrong with the request...")
        print(e)
        print(response)
        print(response.text)


def name_test_qai(segment: str, name: str, meta: dict = None):
    """
    For testing qai services via the k8s proxy (kubectl proxy)

    usage:
    name_test_qai("testing string", "gec")

    appropriate meta will be injected based on service name,
    you can override this by using the meta kwarg
    """
    if meta is None:
        meta = __metas.get(name, {})
    url = f"http://127.0.0.1:8001/api/v1/namespaces/default/services/{name}:80/proxy/v2"
    return test_qai(segment, meta, url)


def test_grammar_script(rules: str, json: bool = False, show: bool = False):
    """call grammar-script in k8s, get json from yaml-like or file
    params:
        - rules: a string of yaml-like grammarScript rules, or path to such a file
        - json: whether to return JSON format (alternative is python dict)
        - show: whether to print or return (True means print, don't return)
    """
    if "rule:" not in rules:
        if os.path.isfile(rules):
            with open(rules) as f:
                rules = "".join([l for l in f])
        else:
            raise ValueError(
                "argument must be grammar-script rules string or path to such a file"
            )
    url = "http://127.0.0.1:8001/api/v1/namespaces/default/services/grammar-script:80/proxy/"
    response = requests.post(url, json={"rules": rules})
    payload = response.json()
    # does anyone need tests?
    md, tests = payload.get("dict", {}), payload.get("tests", {})
    if json:
        if show:
            print(__mod_json.dumps(md, indent=2))
            return
        else:
            return __mod_json.dumps(md)
    if show:
        pprint.pprint(md)
        return
    else:
        return md



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['qai']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.26.0,<3.0.0', 'sanic>=21.3.4,<22.0.0']

setup_kwargs = {
    'name': 'qai',
    'version': '6.0.0rc4',
    'description': 'Helper functions and classes for interacting with the rest of the Writer platform',
    'long_description': '# library.qai.utilities\n\nHelper functions and classes for interacting with the rest of the Writer platform. The main components are:\n\n- `qai.issues.make_issue`: makes a dictionary that conforms to Writer platform standards\n- `qai.spacy_factor.SpacyFactor`: a helper class which turns a spaCy Span into an issue\n- `qai.server.QRest`: a helper class which builds a Sanic REST server for you\n- `qai.validation.Validator`: a simple validator class that can be used to skip segments without processing them, e.g. for being empty or having HTML. You would only want to import this yourself if you aren\'t using `QRest`\n\nIf you are building a Sanic app without using `QRest`, you still may be interested in the following middleware:\n\n- `qai.server.set_valid_segments`: stores valid segments to `request.ctx.segments` and also sets `request.ctx.info`\n- `qai.server.postprocess_issues`: changes the string indexing to be utf-16 based, and filters no ops\n\n(See GitHub history for older docs - QAI used to do a lot more!)\n\n## Testing QAI services\n\nWith QAI >=6.0.0, it is easy to test a QAI-based service running in k8s, first start the k8s proxy with `$kubectl proxy` and then `$ python` (or, if you are fancy, `$ python -m IPython`)\n\n```python\nfrom qai.testing import *\n\nname_test_qai("testing string", "service name")\n```\n\nThat\'s all! The meta is inferred from the service name.\n\nYou can also test a service running on localhost port 5000 by:\n\n```python\nfrom qai.testing import *\n\ntest_qai("testing string", url=5000)\n# or to imply meta by name\ntest_qai("testing string", "service name", url=5000)\n# or to explicitly specify meta\ntest_qai("testing string", meta={"some key": val}, url=5000)\n```\n\n## Upgrading to v5\n\n- `Analyzer` class must now be callable. It will be passed `(segment: str, meta: dict, all_info: dict)` - `segment` is the string to analyze, `meta` is the `meta` object that was sent, or `{}` if none was sent, and `all_info` is the entire payload the server received - in case you need access to clientID or something. Feel free to define `def __call__(self, segment: str, meta: dict, _)` if you don\'t expect to need the request.\n- `QRest` can be passed a Sanic app, or be passed a dictionary which maps issue types to categories (in addition to the default behavior). This is useful for services that handle multiple categories, for which the default behavior doesn\'t work.\n- QAI has a simpler structure, so all imports look different\n- Configs, Strings, Storage, and Document are gone. The later 2 because they aren\'t needed anymore, the former 2 because you should manage that yourself. Whitelisting is also gone - just don\'t make the issues if you don\'t want them.\n- All issues are created in the v2 format (meaning, the format we switched to after new segmentation - defined [here](https://writerai.atlassian.net/wiki/spaces/HOME/pages/2115928140/NLP+Services+API+Contract+Meta))\n- By default, issue `from` and `until` keys are now based on UTF-16 indexing, to make things easier for JS. We add `_from_p` and `_until_p` keys for debugging, which are the Python string indexes. This happens as response middleware in QRest.\n\n## Usage\n\nYou can explicitly create a REST connection like this:\n\n```python\nfrom app import Analyzer\n\nfrom qai.server import QRest\n\n\n# setting the category / service name does nothing\n# we use the category passed on the request\ncategory = \'service_name\'\nhost = \'0.0.0.0\'\nport = 5000\n\n\nif __name__ == \'__main__\':\n    analyzer = Analyzer()\n    rest_connection = QRest(\n      analyzer,\n      category=category,\n      host=host,\n      port=port\n    )\n    # create a blocking connection:\n    rest_connection.connect()\n```\n\nThe above will create *as many workers as you have cores.* This is great, sometimes. For example, there is a known bug where AutoML crashes if you are using more than one worker. So pass `workers=1` if this happens\n\nThere is also a helper class for turning spaCy `Span`s into issues the rest of the platform can process:\n\n```python\nfrom spacy.tokens import Span\nfrom qai.spacy_factor import SpacyFactor\n\n\nMyFactor = SpacyFactor(\n    "subject_object_verb_spacing",\n    "Keep the subject, verb, and object of a sentence close together to help the reader understand the sentence."\n)\n\nSpan.set_extension("score", default=0)\nSpan.set_extension("suggestions", default=[])\n\ndoc = nlp("Holders of the Class A and Class B-1 certificates will be entitled to receive on each Payment Date, to the extent monies are available therefor (but not more than the Class A Certificate Balance or Class B-1 Certificate Balance then outstanding), a distribution.")\nscore = analyze(doc)\nif score is not None:\n    span = Span(doc, 0, len(doc))  # or whichever tokens/spans are the issue (don\'t have to worry about character indexes)\n    span._.score = score\n    span._.suggestions = get_suggestions(doc)\n    issue = MyFactor(span)\n```\n\n## Installation\n\n`pip install qai` or `poetry add qai`\n\n## Testing\n\nSee Confluence for docs on input format expectations.\n\n`scripts/test_qai.sh` has some helpful testing functions.\n\n### CI/CD\n\nGitHub Actions will push to PyPi when you merge into the `main` branch.\n\n### License\n\nThis software is not licensed. If you do not work at Writer, you are not legally allowed to use it. Also, it\'s just helper functions that really won\'t help you. If something in it does look interesting, and you would like access or our help, open an issue.\n',
    'author': 'Sam Havens',
    'author_email': 'sam.havens@writer.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import re
from typing import Any, Dict, List


def filter_no_ops(segment: str, issues: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    filter issues to drop 0-distance edits
    :param issues: issues to filter
    :return: filter issues
    """
    filtered_issues = []
    for issue in issues:
        if issue["suggestions"]:
            suggestions = []
            for suggestion in issue["suggestions"]:
                # different keys because of utf-16 indexing
                # which means from and until indexes aren't usable in python
                f = issue.get("_from_p", None)
                if not f:
                    f = issue.get("from")
                u = issue.get("_until_p", None)
                if not u:
                    u = issue.get("until")
                if (segment[f:u]) == suggestion:
                    continue
                suggestions.append(suggestion)

            if suggestions:
                issue["suggestions"] = suggestions
                filtered_issues.append(issue)
        else:  # some services don't offer suggestions, so no need to check
            filtered_issues.append(issue)
    return filtered_issues


def fix_extra_space(segment: str, issues: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    modify issues in which a deletion would cause a double space
    :param segment: the segment with issues
    :param issues: issues to potentially modify
    :return: modified issues
    """
    modified_issues = []
    for issue in issues:
        # we only care about pure deletes
        # which for us means one suggestion that is the empty string
        # possibly want to cover impure deletes, but gets much harder
        if len(issue["suggestions"]) == 1 and issue["suggestions"][0] == "":
            # different keys because of utf-16 indexing
            # which means from and until indexes aren't usable in python
            f = issue.get("_from_p", None)
            if not f:
                f = issue.get("from")
            u = issue.get("_until_p", None)
            if not u:
                u = issue.get("until")
            # check if either side has spaces
            causes_double_space = (
                f > 0
                and segment[f - 1] == " "
                and len(segment) > u
                and segment[u] == " "
            )
            if causes_double_space:
                # we have a double space issue
                # so extend forward (which is always possible)
                issue["until"] += 1
                if issue.get("_until_p", None) is not None:
                    issue["_until_p"] += 1
            # check if issue is preceded by space and succeeded by SEP
            causes_space_sep = (
                f > 0
                and segment[f - 1] == " "
                and len(segment) == u + 1
                and segment[u] in [".", "?", "!", ",", ";", ":"]
            )
            if causes_space_sep:
                # extend back one char (always possible) to catch extra space
                issue["from"] -= 1
                if issue.get("_from_p", None) is not None:
                    issue["_from_p"] -= 1

        # don't touch most issues
        modified_issues.append(issue)
    return modified_issues


class Validator:
    html_pattern = re.compile("<.*?>")
    non_letter_pattern = re.compile("[^a-zA-Z'\-\s]")

    def __init__(self, ignore_html=True, ignore_token_fraction=0.5):
        self.ignore_html = ignore_html
        self.acceptable_fraction_of_ignorable_tokens = ignore_token_fraction

    def _has_html(self, segment: str) -> bool:
        return self.html_pattern.match(segment) != None

    def _has_non_letter(self, segment: str):
        return self.non_letter_pattern.match(segment) != None

    def _is_empty(self, segment: str):
        return len(segment.strip()) == 0

    def _is_unacceptable(self, segment: str):
        if self._is_empty(segment):
            print("segment is empty")
            return True
        if self.ignore_html and self._has_html(segment):
            print("segment has HTML")
            return True
        return False

    def _is_ignored_tokens(self, token: str) -> bool:
        return self._has_non_letter(token)

    def _has_too_many_ignore_tokens(self, segment: str) -> bool:
        tokens = segment.strip().split(" ")
        token_length = len(tokens)
        ignored_tokens = [t for t in tokens if self._is_ignored_tokens(t)]
        ignored_length = len(ignored_tokens)
        if token_length == 0:
            return False
        else:
            return (
                ignored_length / token_length
                > self.acceptable_fraction_of_ignorable_tokens
            )

    def __call__(self, segment: str):
        if self._is_unacceptable(segment):
            return False
        if self._has_too_many_ignore_tokens(segment):
            print(f"too many non-letter tokens in {segment}")
            return False
        return True
